# Generated by Django 4.0.6 on 2022-07-21 19:27

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uzcloud_billing.utils


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='BillingAccount',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('account_number', models.CharField(default=uzcloud_billing.utils.generate_account_number, max_length=255, unique=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='billing_account', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



import requests
import jwt
import json

from django.conf import settings

import uzcloud_billing.services as services

from .decorators import auth_required


class Singleton(type):
    _instances = {}

    def __call__(self, *args, **kwargs):
        if self not in self._instances:
            self._instances[self] = super(Singleton, self).__call__(*args, **kwargs)
        return self._instances[self]


class UzcloudBilling(metaclass=Singleton):
    AUTH_URL = settings.UZCLOUD_BILLING["AUTH_URL"]
    BILLING_BASE_URL = settings.UZCLOUD_BILLING["BASE_URL"]
    ADD_ACCOUNT_URL = f"{BILLING_BASE_URL}/api/Account/AddAccount"
    GET_BALANCE_URL = f"{BILLING_BASE_URL}/api/Balance/GetBalance"
    MAKE_INVOICE_URL = f"{BILLING_BASE_URL}/api/Invoice/MakeInvoice"

    AUTH_TOKEN = None
    REQUEST_CONFIG = {"verify": False}

    def authorize(self):
        response = requests.post(
            url=self.AUTH_URL,
            data={"grant_type": "client_credentials"},
            auth=(
                settings.UZCLOUD_BILLING["CLIENT_ID"],
                settings.UZCLOUD_BILLING["CLIENT_SECRET"],
            ),
            **self.REQUEST_CONFIG,
        )
        if response.status_code != 200:
            raise ValueError(response.content)
        self.AUTH_TOKEN = response.json()["access_token"]
        self.DECODED = jwt.decode(self.AUTH_TOKEN, options={"verify_signature": False})

    @auth_required
    def add_account(self, personType: int = 1):
        payload = {"personType": personType}
        response = requests.post(
            url=self.ADD_ACCOUNT_URL,
            data=json.dumps(payload),
            headers=self.get_headers(),
            **self.REQUEST_CONFIG,
        )
        if response.status_code != 200:
            raise ValueError(response.content)
        return response.json()

    @auth_required
    def get_balance(self, account_number: str):
        response = requests.get(
            url=self.GET_BALANCE_URL,
            params={"accountNumber": account_number},
            headers=self.get_headers(),
            **self.REQUEST_CONFIG,
        )
        if response.status_code != 200:
            raise ValueError(response.content)
        return response.json()["Amount"]

    @auth_required
    def make_invoice(self, account_number: str, amount: float, reason: str):
        payload = {
            "AccountNumber": account_number,
            "Amount": amount,
            "Reason": reason,
        }
        response = requests.post(
            self.MAKE_INVOICE_URL,
            data=json.dumps(payload),
            headers=self.get_headers(),
            **self.REQUEST_CONFIG,
        )
        if response.status_code != 200:
            raise ValueError(response.json())
        response = response.json()
        services.update_account_balance(
            account_number=account_number,
            balance=response["Balance"],
        )
        return response

    def get_headers(self):
        return {
            "Authorization": f"Bearer {self.AUTH_TOKEN}",
            "Content-Type": "application/json",
        }


uzcloud_service = UzcloudBilling()


def generate_account_number(personType=1):
    account = uzcloud_service.add_account(personType=personType)
    return account.get("AccountNumber")



from django.urls import path
from uzcloud_billing.api.views import BalanceFilledEventView, IdentEventView

urlpatterns = [
    path("payment/send_payment_event/", BalanceFilledEventView.as_view()),
    path("payment/ident/", IdentEventView.as_view()),
]



from django.conf import settings
from django.utils.module_loading import import_string

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from rest_framework.authentication import BasicAuthentication

from uzcloud_billing.models import BillingAccount
from uzcloud_billing.services import update_account_balance
from uzcloud_billing.signals import balance_filled_signal
from .serializers import PaymentEventSerializer, IdentSerializer
from .permissions import IsBillingGroupPermission


class BalanceFilledEventView(APIView):
    authentication_classes = [BasicAuthentication]
    permission_classes = [IsAuthenticated, IsBillingGroupPermission]

    def post(self, request, *args, **kwargs):
        serializer = PaymentEventSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        update_account_balance(
            account_number=serializer.validated_data["AccountNumber"],
            balance=serializer.validated_data["Balance"],
        )
        balance_filled_signal.send(
            sender=BillingAccount,
            instance=serializer.billing_account,
            data=serializer.validated_data,
        )
        return Response()


class IdentEventView(APIView):
    authentication_classes = [BasicAuthentication]
    permission_classes = [IsAuthenticated, IsBillingGroupPermission]

    def post(self, request, *args, **kwargs):
        serializer = IdentSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        response_serializer = import_string(
            settings.UZCLOUD_BILLING["IDENT_RESPONSE_SERIALIZER"]
        )
        return Response(
            response_serializer(instance=serializer.billing_account.user).data
        )



from rest_framework.permissions import BasePermission


class IsBillingGroupPermission(BasePermission):
    def has_permission(self, request, view):
        return request.user.groups.filter(name="Uzcloud_Billing").exists()



from django.core.management.base import BaseCommand
from django.contrib.auth import get_user_model

from uzcloud_billing.models import BillingAccount

User = get_user_model()


class Command(BaseCommand):
    help = "Create Billing Account for Users who don't have billing account"

    def handle(self, *args, **options):
        users_has_no_billing_accounts = User.objects.filter(
            billing_account__isnull=True
        )
        if users_has_no_billing_accounts.exists():
            for user in users_has_no_billing_accounts:
                BillingAccount.objects.create(user=user)
            self.stdout.write(
                self.style.SUCCESS(
                    f"Successfully {users_has_no_billing_accounts.count()} BillingAccounts have been created!"
                )
            )
        else:
            self.stdout.write(
                self.style.NOTICE("All users have already owned billing accounts")
            )



from django.db import models
from django.conf import settings


class BillingAccount(models.Model):
    class AccountTypes(models.IntegerChoices):
        INDIVIDUAL = 1, "Individual"
        ORGANIZATION = 2, "Organization"

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="billing_account",
    )
    account_number = models.CharField(max_length=255, unique=True)
    account_type = models.IntegerField(
        choices=AccountTypes.choices, default=AccountTypes.INDIVIDUAL
    )
    balance = models.DecimalField(max_digits=10, decimal_places=2, default=0)

    def __str__(self):
        return f"{self.user} - {self.account_number}"



import uzcloud_billing.utils as utils
from uzcloud_billing.models import BillingAccount
from uzcloud_billing.signals import payment_completed_signal


def create_billing_account(*, user, personType=1):
    return BillingAccount.objects.create(
        user=user,
        account_number=utils.generate_account_number(personType=personType),
        account_type=personType,
    )


def update_account_balance(*, account_number: str, balance: float):
    billing_account = BillingAccount.objects.get(account_number=account_number)
    billing_account.balance = balance
    billing_account.save()


def make_invoice(account_number: str, amount: float, reason: str, data: dict):
    """
    Eaxample Response :
    {
        "AccountNumber": "AA-000001",
        "ChargeAmount": 3500,
        "Balance": 109429656.176318228,
        "InvoiceId": "c3b9f00e-9c7d-4b36-a8cd-ee561041be93",
        "CreatedAt": "2022-05-07T07:07:00.8624403+00:00"
    }
    """
    response: dict = utils.uzcloud_service.make_invoice(
        account_number=account_number, amount=amount, reason=reason
    )
    data.update(response)
    payment_completed_signal.send(sender=None, data=data)
    return response



from django.contrib import admin

from .models import BillingAccount


@admin.register(BillingAccount)
class BillingAccountAdmin(admin.ModelAdmin):
    list_display = ["user", "account_number", "account_type", "balance"]
    readonly_fields = ["user", "account_number", "account_type", "balance"]
    list_filter = ["account_type"]
    search_fields = ["account_number"]

    def has_delete_permission(self, request, obj=None):
        return False



from django.contrib.auth import get_user_model

from rest_framework import serializers
from rest_framework.exceptions import NotFound

from uzcloud_billing.models import BillingAccount

User = get_user_model()


class BaseAccountNumberSerializer(serializers.Serializer):
    AccountNumber = serializers.CharField(max_length=16, min_length=9)

    def validate_AccountNumber(self, value):
        account = BillingAccount.objects.filter(account_number=value)
        if not account.exists():
            raise NotFound(detail={"error_code": "account_number_not_exist"})
        self.billing_account = account.first()
        return value


class PaymentEventSerializer(BaseAccountNumberSerializer):
    paymentType = serializers.CharField(max_length=20)
    Amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    Balance = serializers.DecimalField(max_digits=10, decimal_places=2)


class IdentSerializer(BaseAccountNumberSerializer):
    pass


class IdentResponseSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ["id", "is_active"]



# Generated by Django 4.0.6 on 2022-07-22 11:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('uzcloud_billing', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='billingaccount',
            name='balance',
            field=models.DecimalField(decimal_places=2, default=0, max_digits=10),
        ),
    ]



from django.utils import timezone
from django.conf import settings


def auth_required(func):
    def wrapper(self, *args, **kwargs):
        if self.AUTH_TOKEN is None:
            self.authorize()
        else:
            now = timezone.now()
            exp = timezone.datetime.fromtimestamp(
                self.DECODED.get("exp", None),
                tz=timezone.utc if settings.USE_TZ else None,
            )
            if now > exp:
                self.authorize()
        return func(self, *args, **kwargs)

    return wrapper



# Generated by Django 4.0.6 on 2022-12-20 09:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('uzcloud_billing', '0002_billingaccount_balance'),
    ]

    operations = [
        migrations.AddField(
            model_name='billingaccount',
            name='account_type',
            field=models.IntegerField(choices=[(1, 'Individual'), (2, 'Organization')], default=1),
        ),
        migrations.AlterField(
            model_name='billingaccount',
            name='account_number',
            field=models.CharField(max_length=255, unique=True),
        ),
    ]
